
/* cmd3.c - functions implementing score processing statements */

#include "main.h"
#include "cmd.h"

/* modulate by another p-field */

void	mod_px (int, double, double, double, int, double*);

/* modulate by noise */

void	mod_noise (int, double, double, double, int, double*);

/* modulate by sine wave */

void	mod_sine (int, double, double, double, int, double*, int);

/* modulate by triangle / saw wave */

void	mod_trsw (int, double, double, double, int, double*, int);

/* modulate main function */

void	cmd_modulate (int nr_args, double *arg_list)
{
	int	m, p;
	double	mc, mm, mp;

	if (nr_args < 5) err_msg_s ("missing argument for modulate", NULL);
	p = check_int (arg_list[0], 0, P_MAX, "invalid p-field number");
	mc = arg_list[1]; mm = arg_list[2]; mp = arg_list[3];
	m = check_int (arg_list[4], 0, 5, "invalid modulation type");
	nr_args -= 4; arg_list += 4;	/* argument list will point to	*/
	switch (m) {			/* modulation type		*/
		case 0:	mod_px (p, mc, mm, mp, nr_args, arg_list); break;
		case 1:	mod_noise (p, mc, mm, mp, nr_args, arg_list); break;
		case 2:	mod_sine (p, mc, mm, mp, nr_args, arg_list, 0); break;
		case 3:	mod_sine (p, mc, mm, mp, nr_args, arg_list, 1); break;
		case 4:	mod_trsw (p, mc, mm, mp, nr_args, arg_list, 0); break;
		case 5:	mod_trsw (p, mc, mm, mp, nr_args, arg_list, 1); break;
	}
}

/* modulate by another p-field */

void	mod_px (int p, double mc, double mm, double mp,
		int nr_args, double *arg_list)
{
	int	e, p2;
	double	x, y;

	if (nr_args < 2) err_msg_s ("missing argument for modulate", NULL);
	if (nr_args > 2) err_msg_s ("too many arguments for modulate", NULL);
	p2 = check_int (arg_list[1], 0, P_MAX, "invalid p-field number");
	e = -1; while ((e = next_event (e)) >= 0) {
		if (read_pfield (&x, e, p)) continue;
		if (read_pfield (&y, e, p2)) y = 0.0;
		x = (x * mc) + (y * mm) + (x * y * mp);
		store_pfield (x, e, p);
	}
}

/* modulate by noise */

void	mod_noise (int p, double mc, double mm, double mp,
		int nr_args, double *arg_list)
{
	int	e;
	double	x, y, pw, s;

	if (nr_args < 1) err_msg_s ("missing argument for modulate", NULL);
	if (nr_args > 2) err_msg_s ("too many arguments for modulate", NULL);
	pw = 0.0; if (nr_args > 1) pw = arg_list[1];
	e = -1; while ((e = next_event (e)) >= 0) {
		if (read_pfield (&x, e, p)) continue;
		y = get_rand ();
		s = (y < 0.0 ? -1.0 : 1.0); y = fabs (y);
		if ((pw == 0.0) || (pw == 1.0) || (pw == -1.0)) {
			y *= s;
		} else if (pw < 0.0) {
			y = s * (1.0 - pow (fabs (1.0 - y), -pw));
		} else {
			y = s * pow (y, pw);
		}
		x = (x * mc) + (y * mm) + (x * y * mp);
		store_pfield (x, e, p);
	}
}

/* modulate by sine wave */

void	mod_sine (int p, double mc, double mm, double mp,
		int nr_args, double *arg_list, int integrate)
{
	int	e;
	double	x, y, x0, x2, l, a, b;

	if (nr_args < 4) err_msg_s ("missing argument for modulate", NULL);
	if (nr_args > 4) err_msg_s ("too many arguments for modulate", NULL);
	x0 = arg_list[1]; check_value (x0, 0.0, 1.0, "invalid start phase");
	x2 = arg_list[2]; check_value (x2, 0.0, 1.0, "invalid end phase");
	l = arg_list[3]; check_value (l, 1.0e-6, 1.0e+6, "invalid loop time");
	x0 *= TWOPI_D; x2 *= TWOPI_D;
	if (fabs (x2 - x0) < 1.0e-6) {
		x0 = 0.0; x2 = 1.0; a = b = 0.0;
	} else {
		x = 1.0; y = -1.0;	/* min, max value */
		if ((x0 < (0.5 * PI_D)) && (x2 > (0.5 * PI_D))) y = 1.0;
		if ((x0 > (0.5 * PI_D)) && (x2 < (0.5 * PI_D))) y = 1.0;
		if ((x0 < (1.5 * PI_D)) && (x2 > (1.5 * PI_D))) x = -1.0;
		if ((x0 > (1.5 * PI_D)) && (x2 < (1.5 * PI_D))) x = -1.0;
		if (sin (x0) > y) y = sin (x0); if (sin (x2) > y) y = sin (x2);
		if (sin (x0) < x) x = sin (x0); if (sin (x2) < x) x = sin (x2);
		a = 2.0 / (y - x);	/* peak to peak amplitude = 2.0 */
		b = cos (x2) - cos (x0);
		if (integrate) {
			a *= (l / (x2 - x0));		/* a: scale	*/
		}
		b *= a / (x2 - x0);			/* b: offset	*/
	}
	e = -1; while ((e = next_event (e)) >= 0) {
		if (read_pfield (&x, e, p)) continue;
		if (read_pfield (&y, e, 2)) y = 0.0;	/* get p2 value */
		y = fmod (y, l) / l; if (y < 0.0) y += 1.0;
		y = x0 + (x2 - x0) * y;
		if (integrate) {
			y = a * (cos (x0) - cos (y)) + b * (y - x0);
		} else {
			y = a * sin (y) + b;
		}
		x = (x * mc) + (y * mm) + (x * y * mp);
		store_pfield (x, e, p);
	}
}

/* modulate by triangle / saw wave */

void	mod_trsw (int p, double mc, double mm, double mp,
		int nr_args, double *arg_list, int integrate)
{
	int	e;
	double	x, y, at, l;

	if (nr_args < 3) err_msg_s ("missing argument for modulate", NULL);
	if (nr_args > 3) err_msg_s ("too many arguments for modulate", NULL);
	at = arg_list[1]; check_value (at, 0.0, 1.0, "invalid attack time");
	l = arg_list[2]; check_value (l, 1.0e-6, 1.0e+6, "invalid loop time");
	if (at < 1.0e-6) at = 0.0; if (at > (1.0 - 1.0e-6)) at = 1.0;
	e = -1; while ((e = next_event (e)) >= 0) {
		if (read_pfield (&x, e, p)) continue;
		if (read_pfield (&y, e, 2)) y = 0.0;	/* get p2 value */
		y = fmod (y, l) / l; if (y < 0.0) y += 1.0;
		if ((at == 1.0) || ((at != 0.0) && (y <= at))) {
			y /= at;
			if (integrate) {
				y = (y * y - y) * at * l;
			} else {
				y = 2.0 * y - 1.0;
			}
		} else {
			y = (y - at) / (1.0 - at);
			if (integrate) {
				y = (y - y * y) * (1.0 - at) * l;
			} else {
				y = 1.0 - 2.0 * y;
			}
		}
		x = (x * mc) + (y * mm) + (x * y * mp);
		store_pfield (x, e, p);
	}
}

