
/* cmd_main.h - read and parse score processing statements */

#include "main.h"
#include "cmd.h"

int		cmd_line_num;		/* line number in program file	*/
event_mask	e_masks[N_MASKS];	/* event masks			*/

/* Get a p-field from event list at location event_num, p_num.	*/
/* Return value is: 0: successfully read, -1: invalid p-field	*/
/* number, -2: invalid event number.				*/

/* The event number used by read_pfield, store_pfield and next_event	*/
/* functions is actually the event_list array index multiplied by 2 and	*/
/* 0 or 1 added for note-on or note-off events. Note-on events are the	*/
/* normal score i-statements, with the extra parameter p0 storing start	*/
/* time offset in beats; all parameters can be read and written.	*/
/* Note-off events exist only for i-statements with positive p3 value.	*/
/* In the case of these events, p0 stores time offset in beats to note-	*/
/* off time, and p2 is equal to (p2 + p3) of the corresponding note-on	*/
/* event. All other parameters are read from this note-on event. Only	*/
/* p0 of note-off events can be written, other p-fields are read-only.	*/

int	read_pfield (double *x, int event_num, int p_num)
{
	int	e, n;

	e = event_num >> 1;		/* index to event list	*/
	n = event_num & 1;		/* note-off		*/
	*x = 0.0;			/* clear output		*/
	if ((e < 0) || (e >= nr_events))
		return -2;	/* event number out of range */
	if ((n) && ((event_list[e].nr_pfields < 3)
		    || (event_list[e].p_list[2] <= 0.0)))
		return -2;	/* no note-off (p3 not > 0.0) */
	if ((p_num < 0) || (p_num > event_list[e].nr_pfields))
		return -1;	/* p-field number out of range */
	if (!(p_num)) {
		if (n) {				/* note-off offset */
			*x = event_list[e].end_ofs;
		} else {				/* time offset	   */
			*x = event_list[e].time_ofs;
		}
	} else {
		if ((n) && (p_num == 2)) {		/* note-off time   */
			*x = event_list[e].p_list[1] + event_list[e].p_list[2];
		} else {				/* read p-field    */
			*x = event_list[e].p_list[p_num - 1];
		}
	}

	return 0;
}

/* Store a p-field in event list at location event_num, p_num.	*/
/* Return value is: 0: successfully wrote, -1: invalid p-field	*/
/* number, -2: invalid event number.				*/

int	store_pfield (double x, int event_num, int p_num)
{
	int	e, n;

	e = event_num >> 1;		/* index to event list	*/
	n = event_num & 1;		/* note-off		*/
	if ((e < 0) || (e >= nr_events))
		return -2;	/* event number out of range */
	if ((n) && ((event_list[e].nr_pfields < 3)
		    || (event_list[e].p_list[2] <= 0.0)))
		return -2;	/* no note-off (p3 not > 0.0) */
	if ((p_num < 0) || (p_num > event_list[e].nr_pfields))
		return -1;	/* p-field number out of range */
	if ((n) && (p_num))
		return -1;	/* only p0 of note-off events can be written */
	if (!(p_num)) {
		if (n) {				/* note-off offset */
			event_list[e].end_ofs = x;
		} else {				/* time offset	   */
			event_list[e].time_ofs = x;
		}
	} else {					/* write p-field    */
		event_list[e].p_list[p_num - 1] = x;
	}

	return 0;
}

/* Return the number of next event, or -1 if none was found. */

int	next_event (int event_num)
{
	int	e, i, j;
	double	x;

	e = event_num;
	if (e < -1) return -1;		/* invalid event number */
	while (++e < (nr_events << 1)) {
		j = read_pfield (&x, e, 0);
		if (j == -2) continue;
		i = -1; while (++i < N_MASKS) {		/* check masks */
			if (!(e_masks[i].m_mode))
				continue;		/* mask disabled */
			j = read_pfield (&x, e, e_masks[i].p_num);
			if (j < 0) {
				j = 1;	/* p-field number out of range */
			} else {
				j = 0;
				if (e_masks[i].loop > 0.0)
					x += e_masks[i].loop
					     * ceil ((e_masks[i].l_limit - x)
						     / e_masks[i].loop);
				if ((x < e_masks[i].l_limit)
				    || (x > e_masks[i].u_limit))
					j = 1;
			}
			if (e_masks[i].m_mode == 2) j = !j;
			if (j) break;
		}
		if (i >= N_MASKS) return e;	/* return next event number */
	}

	return -1;	/* no event was found */
}

/* parse score processing statements */

void	cmd_main (void)
{
	char	*stat;		/* string containing current statement	*/
	double	*arg_list;	/* list of arguments			*/
	int	i, j;

	cmd_line_num = 0;
	stat = NULL;
	arg_list = NULL;
	r_seed = 0x40000000UL;			/* initialize random seed */
	do {
		/* read next line from program file */

		i = get_statement (cmd_file, &stat, &arg_list);
		cmd_line_num++;

		if (i < 0) break;		/* end of file	*/
		if (stat == NULL) continue;	/* empty line	*/

		j = -1; while (++j < (int) strlen (stat))	/* convert */
			stat[j] = tolower (stat[j]);		/* case    */

		if (!(strcmp (stat, "pformat"))) {	/* priniting format */
			cmd_pformat (i, arg_list);
		} else if (!(strcmp (stat, "mask"))) {	/* event mask */
			cmd_mask (i, arg_list);
		} else if (!(strcmp (stat, "set"))) {	/* set */
			cmd_set (i, arg_list);
		} else if (!(strcmp (stat, "add"))) {	/* add */
			cmd_add (i, arg_list);
		} else if (!(strcmp (stat, "mul"))) {	/* multiply */
			cmd_mul (i, arg_list);
		} else if (!(strcmp (stat, "pow"))) {	/* pow */
			cmd_pow (i, arg_list);
		} else if (!(strcmp (stat, "quant"))) {	/* quantize */
			cmd_quant (i, arg_list);
		} else if (!(strcmp (stat, "limit"))) {	/* limit */
			cmd_limit (i, arg_list);
		} else if (!(strcmp (stat, "seed"))) {	/* set random seed */
			cmd_seed (i, arg_list);
		} else if (!(strcmp (stat, "modulate"))) {	/* modulate */
			cmd_modulate (i, arg_list);
		} else {
			err_msg_s ("invalid statement", stat);
		}

		/* free memory used by temp variables */

		if (stat != NULL) {
			free (stat); stat = NULL;
		}
		if (arg_list != NULL) {
			free (arg_list); arg_list = NULL;
		}
	} while (1);
}

